from __future__ import annotations

import types
import typing

import pydantic

DictConvertible = typing.Union[typing.Mapping[str, typing.Any], typing.Iterable[tuple[str, typing.Any]]]

try:
    UnionType = types.UnionType
except AttributeError:
    UnionType = typing._UnionGenericAlias  # noqa: SLF001


class Model(pydantic.BaseModel):
    _eq_excluded_fields: typing.ClassVar[set[str]] = set()
    _skip_prefix_fields: typing.ClassVar[dict[str, str] | None] = None
    _skip_sortable_fields: typing.ClassVar[set[str] | None] = None

    @classmethod
    def _pdb_model_fields(cls: type[typing.Self]) -> dict[str, tuple[Model, bool]]:
        ret = {}
        for k, f in cls.model_fields.items():
            if type(f.annotation) is UnionType:
                args = typing.get_args(f.annotation)
                for arg in args:
                    if isinstance(arg, type) and issubclass(arg, Model):
                        ret[k] = (arg, type(None) in args)
                        break
            elif isinstance(f.annotation, type) and issubclass(f.annotation, Model):
                ret[k] = (f.annotation, False)
        return ret

    def __eq__(self, other: object) -> bool:
        """Equality method to support testing."""
        if type(self) is type(other):
            return all(
                getattr(self, field) == getattr(other, field)
                for field in type(self).model_fields
                if field not in self._eq_excluded_fields
            )
        return False

    @classmethod
    def _parse_result(cls: type[typing.Self], result: DictConvertible, *, prefix: str = "") -> dict:
        # Strip prefixes away
        data = {k.replace(f"{prefix}", ""): v for k, v in dict(result).items() if k.startswith(prefix)}

        skip_prefix_map = cls._skip_prefix_fields or {}
        model_fields = cls._pdb_model_fields()

        for model_prefix, config_ in model_fields.items():
            model, optional = config_
            skip_field = skip_prefix_map.get(model_prefix, "id")
            if optional and data.get(f"{model_prefix}__{skip_field}") is None:
                data[model_prefix] = None
            else:
                # Unions
                data[model_prefix] = model.from_result(
                    {
                        k.replace(f"{model_prefix}__", ""): v
                        for k, v in data.items()
                        if k.startswith(f"{model_prefix}__")
                    },
                )

        return data

    @classmethod
    def from_result(cls: type[typing.Self], result: DictConvertible, *, prefix: str = "") -> typing.Self:
        data = cls._parse_result(result, prefix=prefix)
        return cls(**data)

    @classmethod
    def from_results(
        cls: type[typing.Self],
        results: typing.Sequence[DictConvertible],
        *,
        prefix: str = "",
    ) -> list[typing.Self]:
        return [cls.from_result(r, prefix=prefix) for r in results]

    @classmethod
    def as_columns(cls, base_table: str | None = None) -> list[tuple[str, ...]]:
        return list(cls.as_typed_columns(base_table=base_table).keys())

    @classmethod
    def as_typed_columns(cls, base_table: str | None = None) -> dict[tuple[str, ...], type[typing.Any] | None]:
        columns: dict[tuple[str, ...], type[typing.Any] | None] = {}
        model_fields = cls._pdb_model_fields()

        for field, field_data in cls.model_fields.items():
            if field in model_fields:
                for column, annotation in model_fields[field][0].as_typed_columns().items():
                    columns[(field, *column)] = annotation

            elif base_table is None:
                columns[(field,)] = field_data.annotation
            else:
                columns[(base_table, field)] = field_data.annotation

        return columns

    @classmethod
    def sortable_fields(cls, *, top_level: bool = True) -> set[str]:
        fields = set()
        model_fields = cls._pdb_model_fields()
        skipped_fields = cls._skip_sortable_fields or set()

        for field in cls.model_fields:
            if field in skipped_fields:
                continue

            if field in model_fields:
                if top_level:
                    fields.add(field)

                for column in model_fields[field][0].sortable_fields(top_level=False):
                    sortable_field = f"{field}__{column}"
                    if sortable_field in skipped_fields:
                        continue

                    fields.add(sortable_field)
            else:
                fields.add(field)

        return fields



from __future__ import annotations

import typing
from datetime import datetime, timezone

import pytest

from tests.model import ModelA, ModelB, ModelC, ModelD, ModelE


def test_unrelated_models_not_equal():
    m1 = ModelA(id=1, a="x")
    m2 = ModelB(id=1, b="x")
    assert m1 != m2


def test_different_data_models_not_equal():
    m1 = ModelA(id=1, a="y")
    m2 = ModelA(id=1, a="x")
    assert m1 != m2


def test_equivalent_models_equal():
    m1 = ModelA(id=1, a="y")
    m2 = ModelA(id=1, a="y")
    assert m1 == m2


def test_equivalent_models_equal_ignored_field():
    m1 = ModelC(id=1, c="x", updated=datetime(2025, 1, 1, tzinfo=timezone.utc))
    m2 = ModelC(id=1, c="x", updated=datetime(2025, 1, 2, tzinfo=timezone.utc))
    assert m1 == m2


class TestModel:
    def test_model_fields(self):
        assert ModelA._pdb_model_fields() == {}

    def test_as_columns(self):
        columns = ModelA.as_columns()

        assert columns == [("id",), ("a",)]

    def test_as_typed_columns(self):
        columns = ModelA.as_typed_columns()

        assert columns == {("id",): int, ("a",): str}

    def test_from_result(self):
        r = {"id": 1, "a": "x"}
        model = ModelA.from_result(r)

        assert model == ModelA(id=1, a="x")

    def test_from_result_with_prefix(self):
        r = {"xxxid": 1, "xxxa": "x"}
        model = ModelA.from_result(r, prefix="xxx")

        assert model == ModelA(id=1, a="x")

    def test_from_results(self):
        results = [{"id": 1, "a": "x"}]
        models = ModelA.from_results(results)

        assert models == [
            ModelA(id=1, a="x"),
        ]

    def test_from_results_with_prefix(self):
        results = [{"xxxid": 1, "xxxa": "x"}]
        models = ModelA.from_results(results, prefix="xxx")

        assert models == [
            ModelA(id=1, a="x"),
        ]


class TestNestedModel:
    @pytest.mark.parametrize(
        ("model", "expected_fields"),
        [
            (ModelD, {"a": (ModelA, False), "b": (ModelB, True)}),
            (ModelE, {"d": (ModelD, False)}),
        ],
    )
    def test_model_fields(self, model, expected_fields):
        assert model._pdb_model_fields() == expected_fields

    def test_as_columns(self):
        columns = ModelD.as_columns()

        assert columns == [("id",), ("d",), ("a", "id"), ("a", "a"), ("b", "id"), ("b", "b")]

    def test_as_typed_columns(self):
        columns = ModelE.as_typed_columns()

        assert columns == {
            ("id",): typing.Union[int, float],
            ("e",): str,
            ("d", "id"): int,
            ("d", "d"): str,
            ("d", "a", "id"): int,
            ("d", "a", "a"): str,
            ("d", "b", "id"): int,
            ("d", "b", "b"): str,
        }

    def test_from_result(self):
        r = {"id": 1, "d": "x", "a__id": 2, "a__a": "y", "b__id": 3, "b__b": "z"}
        model = ModelD.from_result(r)

        assert model == ModelD(
            id=1,
            d="x",
            a=ModelA(id=2, a="y"),
            b=ModelB(id=3, b="z"),
        )

    def test_from_result_skips_optional(self):
        r = {"id": 1, "d": "x", "a__id": 2, "a__a": "y", "b__id": None, "b__b": None}
        model = ModelD.from_result(r)

        assert model == ModelD(
            id=1,
            d="x",
            a=ModelA(id=2, a="y"),
            b=None,
        )

    def test_multi_layer_nesting(self):
        r = {"id": 0, "e": "w", "d__id": 1, "d__d": "x", "d__a__id": 2, "d__a__a": "y", "d__b__id": 3, "d__b__b": "z"}
        model = ModelE.from_result(r)

        assert model == ModelE(
            id=0,
            e="w",
            d=ModelD(
                id=1,
                d="x",
                a=ModelA(id=2, a="y"),
                b=ModelB(id=3, b="z"),
            ),
        )

    def test_from_results(self):
        results = [{"id": 1, "d": "x", "a__id": 2, "a__a": "y", "b__id": 3, "b__b": "z"}]
        models = ModelD.from_results(results)

        assert models == [
            ModelD(
                id=1,
                d="x",
                a=ModelA(id=2, a="y"),
                b=ModelB(id=3, b="z"),
            ),
        ]

    @pytest.mark.parametrize(
        ("model", "expected_fields"),
        [
            (ModelD, {"id", "d", "a", "a__id", "a__a", "b", "b__id", "b__b"}),
            (ModelE, {"id", "e", "d", "d__id", "d__d", "d__a__a", "d__b__b"}),
        ],
    )
    def test_sortable_fields(self, model, expected_fields):
        assert model.sortable_fields() == expected_fields



import dsnparse
import mysql.connector
import pytest

from tests.model import ModelA, ModelB, ModelD, ModelE


@pytest.fixture
def cursor(mysql_dsn):
    r = dsnparse.parse(mysql_dsn)
    db = mysql.connector.connect(
        username=r.username,
        password=r.password,
        host=r.host,
        port=r.port,
        database=r.paths[0],
    )
    yield db.cursor(dictionary=True)
    db.close()


class TestModel:
    def test_from_result(self, cursor):
        cursor.execute("select 1 as id, 'x' as a")
        r = cursor.fetchone()
        model = ModelA.from_result(r)

        assert model == ModelA(id=1, a="x")

    def test_from_results(self, cursor):
        cursor.execute("select 1 as id, 'x' as a")
        results = cursor.fetchall()
        models = ModelA.from_results(results)

        assert models == [
            ModelA(id=1, a="x"),
        ]


class TestNestedModel:
    def test_from_result(self, cursor):
        cursor.execute("""
        select
            1 as id,
            'x' as d,
            2 as a__id,
            'y' as a__a,
            3 as b__id,
            'z' as b__b
        """)
        r = cursor.fetchone()
        model = ModelD.from_result(r)

        assert model == ModelD(
            id=1,
            d="x",
            a=ModelA(id=2, a="y"),
            b=ModelB(id=3, b="z"),
        )

    def test_from_result_skips_optional(self, cursor):
        cursor.execute("""
        select
            1 as id,
            'x' as d,
            2 as a__id,
            'y' as a__a,
            null as b__id,
            null as b__b
        """)
        r = cursor.fetchone()
        model = ModelD.from_result(r)

        assert model == ModelD(
            id=1,
            d="x",
            a=ModelA(id=2, a="y"),
            b=None,
        )

    def test_multi_layer_nesting(self, cursor):
        cursor.execute("""
        select
            0 as id,
            'w' as e,
            1 as d__id,
            'x' as d__d,
            2 as d__a__id,
            'y' as d__a__a,
            3 as d__b__id,
            'z' as d__b__b
        """)
        r = cursor.fetchone()
        model = ModelE.from_result(r)

        assert model == ModelE(
            id=0,
            e="w",
            d=ModelD(
                id=1,
                d="x",
                a=ModelA(id=2, a="y"),
                b=ModelB(id=3, b="z"),
            ),
        )

    def test_from_results(self, cursor):
        cursor.execute("""
        select
            1 as id,
            'x' as d,
            2 as a__id,
            'y' as a__a,
            3 as b__id,
            'z' as b__b
        """)
        results = cursor.fetchall()
        models = ModelD.from_results(results)

        assert models == [
            ModelD(
                id=1,
                d="x",
                a=ModelA(id=2, a="y"),
                b=ModelB(id=3, b="z"),
            ),
        ]
